package org.bitbucket.bradleysmithllc.etlagent.informatica.provider;

import com.fasterxml.jackson.databind.JsonNode;
import org.bitbucket.bradleysmithllc.etlagent.informatica.dto.RemoteFile;
import org.bitbucket.bradleysmithllc.etlagent.informatica.dto.json.*;
import org.bitbucket.bradleysmithllc.etlunit.RuntimeSupport;
import org.bitbucket.bradleysmithllc.etlunit.io.FileBuilder;
import org.bitbucket.bradleysmithllc.etlunit.util.IOUtils;
import org.bitbucket.bradleysmithllc.etlunit.util.Incomplete;
import org.bitbucket.bradleysmithllc.etlunit.util.VelocityUtil;
import org.bitbucket.bradleysmithllc.json.validator.JsonUtils;

import java.io.File;
import java.net.InetAddress;
import java.net.UnknownHostException;
import java.util.*;

public class InformaticaDomain {
	private final String domainName;
	private final String clientVersion;
	private final String connectivityHost;
	private final int connectivityPort;
	private final String userName;
	private final String passwordEncrypted;
	private final String password;
	private final String securityDomain;

	private final File domainFile;

	private final File informaticaBinDirectory;
	private final File workingRoot;

	private final RuntimeSupport runtimeSupport;

	private final InformaticaRepository defaultRepository;
	private final InformaticaWebServicesHub defaultWebServicesHub;
	private final DomainsProperty.ClientType clientType;

	private int numRequestsSent = 0;

	private final InformaticaIntegrationService defaultIntegrationService;

	private final List<InformaticaRepository> repositories = new ArrayList<InformaticaRepository>();

	private final String agentHost;
	private final int agentPort;

	@Incomplete(explanation = "Implement domains.infa as a user-supplied file")
	public InformaticaDomain(String name, DomainsProperty value, DomainDefaults domainDefaults, RuntimeSupport runtimeSupport, File domainsFile) {
		domainName = name;

		String cv = value.getClientVersion();

		if (cv == null && domainDefaults != null)
		{
			cv = domainDefaults.getClientVersion();
		}

		checkRequiredAttribute("client-version", cv);

		clientVersion = cv;

		String ch  = value.getConnectivityHost();

		if (ch == null && domainDefaults != null)
		{
			ch = domainDefaults.getConnectivityHost();
		}

		connectivityHost = ch;

		String sd = value.getSecurityDomain();

		if (sd == null && domainDefaults != null)
		{
			sd = domainDefaults.getSecurityDomain();
		}

		securityDomain = sd;

		this.runtimeSupport = runtimeSupport;

		DomainsProperty.ClientType cltype = value.getClientType();

		if (cltype == null && domainDefaults != null)
		{
			cltype = domainDefaults.getClientType();
		}

		clientType = resolveType(cltype);

		switch (clientType) {
			case AGENT:
				// agent is a required property now
				Agent agentNode = value.getAgent();

				if (agentNode == null && domainDefaults != null)
				{
					agentNode = domainDefaults.getAgent();
				}

				if (agentNode == null) {
					throw new IllegalArgumentException("agent property is required when client type is 'agent'");
				}

				agentHost = agentNode.getHostName();

				if (agentHost == null) {
					throw new IllegalArgumentException("agent object requires host-name");
				}

				Long jsonNode = agentNode.getHostPort();

				if (jsonNode != null) {
					agentPort = jsonNode.intValue();
				} else {
					agentPort = 8188;
				}
				break;
			case MOCK:
			case LOCAL:
			default:
				agentHost = null;
				agentPort = 0;
				break;
		}

		String root = value.getWorkingRoot();

		if (root == null && domainDefaults != null)
		{
			root = domainDefaults.getWorkingRoot();
		}

		checkRequiredAttribute("working-root", root);

		workingRoot = new File(root);

		String binRoot = value.getInformaticaBinDirectory();

		if (binRoot == null && domainDefaults != null)
		{
			binRoot = domainDefaults.getInformaticaBinDirectory();
		}

		if (binRoot != null) {
			informaticaBinDirectory = new File(binRoot);
		} else {
			informaticaBinDirectory = null;
		}

		Long port = value.getConnectivityPort();

		if (port == null && domainDefaults != null)
		{
			port = domainDefaults.getConnectivityPort();
		}

		if (port != null) {
			connectivityPort = port.intValue();
		} else {
			connectivityPort = -1;
		}

		String un = value.getUsername();

		if (un == null && domainDefaults != null)
		{
			un = domainDefaults.getUsername();
		}

		userName = un;

		checkRequiredAttribute("username", userName);

		String pw = value.getPasswordEncrypted();

		if (pw == null && domainDefaults != null)
		{
			pw = domainDefaults.getPasswordEncrypted();
		}

		passwordEncrypted = pw;

		checkRequiredAttribute("password-encrypted", passwordEncrypted);

		pw = value.getPassword();

		if (pw == null && domainDefaults != null)
		{
			pw = domainDefaults.getPassword();
		}

		password = pw;

		Repositories erepositories = value.getRepositories();

		if (erepositories == null) {
			throw new IllegalArgumentException("Informatica domain missing repositories");
		}

		Map<String, RepositoriesProperty> it = erepositories.getAdditionalProperties();

		for (Map.Entry<String, RepositoriesProperty> entry : it.entrySet()) {
			InformaticaRepository repository = new InformaticaRepository(this, entry.getKey(), runtimeSupport);

			if (getRepository(entry.getKey()) != null) {
				throw new IllegalArgumentException("Duplicate repository declaration: " + entry.getKey());
			}

			repositories.add(repository);

			// load integration services
			RepositoriesProperty repositoriesProperty = entry.getValue();
			List<String> intSvcs = repositoriesProperty.getIntegrationServices();

			if (intSvcs != null) {
				for (String service : intSvcs) {
					if (getIntegrationService(service) != null) {
						throw new IllegalArgumentException("Integration service " + service + " specified more than once");
					}

					repository.getIntegrationServices().add(new InformaticaIntegrationService(service, repository, runtimeSupport));
				}
			}

			// load integration services
			WebServicesHubs webSvcs = repositoriesProperty.getWebServicesHubs();

			if (webSvcs != null) {
				for (Map.Entry<String, WebServicesHubsProperty> webSvc : webSvcs.getAdditionalProperties().entrySet()) {
					if (getWebServicesHub(webSvc.getKey()) != null) {
						throw new IllegalArgumentException("Web services hub " + webSvc.getKey() + " specified more than once");
					}

					WebServicesHubsProperty webSvcHubValue = webSvc.getValue();
					String hostName = webSvcHubValue.getHostName();

					// use the connectivity host name if provided
					if (hostName == null && connectivityHost != null)
					{
						hostName = connectivityHost;
					}

					// if it is still null, use the local host name
					if (hostName == null) {
						try {
							hostName = InetAddress.getLocalHost().getHostAddress();
						} catch (UnknownHostException e) {
							throw new RuntimeException("Could not determine local host name", e);
						}
					}

					Long hostPort = webSvcHubValue.getHostPort();

					if (hostPort == null) {
						hostPort = new Long(7333L);
					}

					repository.getWebServicesHubs().add(new InformaticaWebServicesHub(webSvc.getKey(), repository, hostName, hostPort.intValue(), runtimeSupport));
				}
			}
		}

		if (repositories.size() == 0) {
			throw new IllegalArgumentException("Informatica domain repositories element is empty");
		}

		String defRepo = value.getDefaultRepository();

		InformaticaRepository repo = repositories.get(0);

		if (defRepo != null) {
			repo = getRepository(defRepo);

			if (repo == null) {
				throw new IllegalArgumentException("Default repository does not exist in domain: " + defRepo);
			}
		}

		defaultRepository = repo;

		String defIs = value.getDefaultIntegrationService();

		InformaticaIntegrationService dis = getDefaultRepository().getDefaultIntegrationService();

		if (defIs != null) {
			dis = getIntegrationService(defIs);

			if (dis == null) {
				throw new IllegalArgumentException("Default integration service not found in domain: " + defIs);
			}
		}

		defaultIntegrationService = dis;

		String defIsh = value.getDefaultWebServicesHub();

		InformaticaWebServicesHub dish = getDefaultRepository().getDefaultWebServicesHub();

		if (defIsh != null) {
			dish = getWebServicesHub(defIsh);

			if (dish == null) {
				throw new IllegalArgumentException("Default web services hub not found in domain: " + defIsh);
			}
		}

		defaultWebServicesHub = dish;

		if (getConnectivityHost() != null) {
			String template = null;
			try {
				template = IOUtils.readURLToString(getClass().getResource("/domains.infa.vm"));

				Map<String, String> bean = new HashMap<String, String>();
				bean.put("domainName", getDomainName());
				bean.put("connectivityHost", getConnectivityHost());
				bean.put("connectivityPort", String.valueOf(getConnectivityPort()));

				String vel = VelocityUtil.writeTemplate(template, bean);

				domainFile = runtimeSupport.createTempFile("domains.infa." + getDomainName());

				IOUtils.writeBufferToFile(domainFile, new StringBuffer(vel));
			} catch (Exception e) {
				throw new IllegalStateException(e);
			}
		} else {
			domainFile = domainsFile;
		}
	}

	private void checkRequiredAttribute(String s, String clientVersion) {
		if (clientVersion == null) {
			throw new IllegalArgumentException("Missing required attribute: '" + s + "'");
		}
	}

	private DomainsProperty.ClientType resolveType(DomainsProperty.ClientType name) {
		if (name == null) {
			return DomainsProperty.ClientType.LOCAL;
		}

		return name;
	}

	public InformaticaWebServicesHub getWebServicesHub(String name) {
		for (InformaticaRepository irepo : repositories) {
			for (InformaticaWebServicesHub iis : irepo.getWebServicesHubs()) {
				if (iis.getName().equals(name)) {
					return iis;
				}
			}
		}

		return null;
	}

	public String getSecurityDomain() {
		return securityDomain;
	}

	public File getInformaticaBinDirectory() {
		return informaticaBinDirectory;
	}

	private String getOptionalAttribute(String s, JsonNode value) {
		return getAttribute(s, value, false);
	}

	private String getRequiredAttribute(String s, JsonNode value) {
		return getAttribute(s, value, true);
	}

	private String getAttribute(String s, JsonNode value, boolean required) {
		JsonNode res = JsonUtils.query(value, s);

		if (required && res == null) {
			throw new IllegalArgumentException("Informatica domain missing attribute: " + s);
		}

		return res == null ? null : res.asText();
	}

	public String getDomainName() {
		return domainName;
	}

	public String getClientVersion() {
		return clientVersion;
	}

	public String getConnectivityHost() {
		return connectivityHost;
	}

	public int getConnectivityPort() {
		return connectivityPort;
	}

	public String getUserName() {
		return userName;
	}

	public String getPasswordEncrypted() {
		return passwordEncrypted;
	}

	public String getPassword() {
		return password;
	}

	public InformaticaRepository getRepository(String name) {
		for (InformaticaRepository repo : repositories) {
			if (repo.getRepositoryName().equals(name)) {
				return repo;
			}
		}

		return null;
	}

	public List<InformaticaRepository> getRepositories() {
		return repositories;
	}

	public InformaticaRepository getDefaultRepository() {
		return defaultRepository;
	}

	public File getWorkingRoot() {
		return workingRoot;
	}

	public File getLocalizedWorkingRoot() {
		return new FileBuilder(workingRoot).subdir(runtimeSupport.getProjectUID()).file();
	}

	public InformaticaIntegrationService getDefaultIntegrationService() {
		return defaultIntegrationService;
	}

	public InformaticaWebServicesHub getDefaultWebServicesHub() {
		return defaultWebServicesHub;
	}

	public InformaticaIntegrationService getIntegrationService(String isName) {
		for (InformaticaRepository irepo : repositories) {
			for (InformaticaIntegrationService iis : irepo.getIntegrationServices()) {
				if (iis.getIntegrationServiceName().equals(isName)) {
					return iis;
				}
			}
		}

		return null;
	}

	public boolean containsIntegrationService(String isName) {
		return getIntegrationService(isName) != null;
	}

	public File getDomainFile() {
		if (domainFile == null) {
			throw new UnsupportedOperationException("Domain file not supported yet");
		}

		return domainFile;
	}

	public DomainsProperty.ClientType getClientType() {
		return clientType;
	}

	public String getAgentHost() {
		return agentHost;
	}

	public int getAgentPort() {
		return agentPort;
	}

	public void dispose() {
		for (InformaticaRepository repository : getRepositories()) {
			repository.dispose();
		}
	}

	File getInformaticaBinDir() {
		// get the client version for the binary
		String ver = getClientVersion();
		// construct a path to the power center binaries using a well-known path combined with the version number

		// check for an explicit bin dir, then use the etl unit dir
		if (informaticaBinDirectory != null) {
			return informaticaBinDirectory;
		} else {
			return new FileBuilder(runtimeSupport.getVendorBinaryDir()).subdir("PowerCenter").subdir(ver).subdir("server").subdir("bin").file();
		}
	}

	public File getWorkspaceDir(RemoteFile.fileType type) {
		return new FileBuilder(getLocalizedWorkingRoot()).mkdirs().subdir(type.getDirectoryName()).mkdirs().file();
	}

	public File getTargetFilesDir() {
		return getWorkspaceDir(RemoteFile.fileType.targetFiles);
	}

	public File getBadFilesDir() {
		return getWorkspaceDir(RemoteFile.fileType.badFiles);
	}

	public File getSessionLogDir() {
		return getWorkspaceDir(RemoteFile.fileType.sessionLog);
	}

	public File getWorkflowLogDir() {
		return getWorkspaceDir(RemoteFile.fileType.workflowLog);
	}

	public File getParameterFilesDir() {
		return getWorkspaceDir(RemoteFile.fileType.parameterFiles);
	}

	public File getSourceFilesDir() {
		return getWorkspaceDir(RemoteFile.fileType.sourceFiles);
	}

	public File getLookupFilesDir() {
		return getWorkspaceDir(RemoteFile.fileType.lookupFiles);
	}

	void request() {
		numRequestsSent++;
	}

	public int getNumRequestsSent() {
		return numRequestsSent;
	}

	@Override
	public String toString() {
		return "InformaticaDomain{" +
				"domainName='" + domainName + '\'' +
				'}';
	}
}